# Import all converters at the top to make sure they are registered first
from .converters import *  # noqa: F403, I001

from .compiler import optimize, optimize_single
from .fields import (
    DjangoConnectionField,
    DjangoListField,
    RelatedField,
    AnnotatedField,
    MultiField,
    ManuallyOptimizedField,
)
from .types import DjangoObjectType

__all__ = [
    "AnnotatedField",
    "DjangoConnectionField",
    "DjangoListField",
    "DjangoObjectType",
    "ManuallyOptimizedField",
    "MultiField",
    "RelatedField",
    "optimize",
    "optimize_single",
]



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

import graphene
from graphene.utils.str_converters import to_snake_case
from graphene_django.settings import graphene_settings
from graphene_django.utils import DJANGO_FILTER_INSTALLED
from graphql import FieldNode, get_argument_values
from graphql.execution.execute import get_field_def

from .ast import GrapheneType, GraphQLASTWalker, Selections, get_underlying_type, is_connection, is_node
from .typing import GQLInfo, GraphQLFilterInfo, ToManyField, ToOneField
from .utils import swappable_by_subclassing

if TYPE_CHECKING:
    from django.db.models import Model
    from graphene.types.definitions import GrapheneObjectType

    from .typing import Any, Optional


__all__ = [
    "get_filter_info",
]


def get_filter_info(info: GQLInfo, model: type[Model]) -> GraphQLFilterInfo:
    """Compile filter information included in the GraphQL query."""
    compiler = FilterInfoCompiler(info, model)
    compiler.run()
    # Return the compiled filter info, or an empty dict if there is no filter info.
    name = getattr(info.field_nodes[0].alias, "value", None) or to_snake_case(info.field_name)
    return compiler.filter_info.get(name, {})


@swappable_by_subclassing
class FilterInfoCompiler(GraphQLASTWalker):
    """Class for compiling filtering information from a GraphQL query."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.filter_info: dict[str, GraphQLFilterInfo] = {}
        super().__init__(*args, **kwargs)

    def add_filter_info(self, parent_type: GrapheneObjectType, field_node: FieldNode) -> None:
        """
        Adds filter info for a field.
        Called for all field nodes that match a database relation.

        :param parent_type: Parent object type.
        :param field_node: FieldNode for the relation.
        """
        graphql_field = get_field_def(self.info.schema, parent_type, field_node)
        graphene_type = get_underlying_type(graphql_field.type)

        field_name = self.get_field_name(field_node)
        orig_field_name = to_snake_case(field_node.name.value)
        filters = get_argument_values(graphql_field, field_node, self.info.variable_values)

        is_node_ = is_node(graphql_field)
        is_connection_ = is_connection(graphene_type)

        # Find the field-specific limit, or use the default limit.
        max_limit: Optional[int] = getattr(
            getattr(parent_type.graphene_type, orig_field_name, None),
            "max_limit",
            graphene_settings.RELAY_CONNECTION_MAX_LIMIT,
        )

        self.filter_info[field_name] = GraphQLFilterInfo(
            name=graphene_type.name,
            # If the field is a relay node field, its `id` field should not be counted as a filter.
            filters={} if is_node_ else filters,
            children={},
            filterset_class=None,
            is_connection=is_connection_,
            is_node=is_node_,
            max_limit=max_limit,
        )

        if DJANGO_FILTER_INSTALLED and hasattr(graphene_type, "graphene_type"):
            object_type = graphene_type.graphene_type
            if is_connection_:
                object_type = object_type._meta.node

            self.filter_info[field_name]["filterset_class"] = getattr(object_type._meta, "filterset_class", None)

    def handle_selections(self, field_type: GrapheneType, selections: Selections) -> None:
        super().handle_selections(field_type, selections)
        # Remove filter info that do not have filters or children.
        # Preserve filter info for connections so that default nested limiting can be applied.
        for name in list(self.filter_info):
            info = self.filter_info[name]
            if not (info["filters"] or info["children"] or info["is_connection"]):
                del self.filter_info[name]

    def handle_query_class(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        self.add_filter_info(field_type, field_node)
        with self.child_filter_info(field_node):
            super().handle_query_class(field_type, field_node)

    def handle_custom_field(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)
        field = getattr(field_type.graphene_type, field_name, None)
        if isinstance(field, graphene.Field):
            self.add_filter_info(field_type, field_node)

    def handle_to_one_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
        related_model: type[Model] | None,
    ) -> None:
        self.add_filter_info(field_type, field_node)
        with self.child_filter_info(field_node):
            return super().handle_to_one_field(field_type, field_node, related_field, related_model)

    def handle_to_many_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
        related_model: type[Model] | None,
    ) -> None:
        self.add_filter_info(field_type, field_node)
        with self.child_filter_info(field_node):
            return super().handle_to_many_field(field_type, field_node, related_field, related_model)

    @contextlib.contextmanager
    def child_filter_info(self, field_node: FieldNode) -> None:
        field_name = self.get_field_name(field_node)
        arguments: dict[str, GraphQLFilterInfo] = {}
        orig_arguments = self.filter_info
        try:
            self.filter_info = arguments
            yield
        finally:
            self.filter_info = orig_arguments
            if arguments:
                self.filter_info[field_name]["children"] = arguments



from __future__ import annotations

from typing import TYPE_CHECKING

import graphene
import graphene_django
from django_filters.constants import ALL_FIELDS
from graphene_django.utils import is_valid_django_model

from .compiler import optimize_single
from .settings import optimizer_settings
from .typing import PK, OptimizedDjangoOptions

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet

    from .optimizer import QueryOptimizer
    from .typing import Any, GQLInfo, Literal, Optional, TModel, Union


__all__ = [
    "DjangoObjectType",
]


class DjangoObjectType(graphene_django.types.DjangoObjectType):
    """DjangoObjectType that automatically optimizes its queryset."""

    _meta: OptimizedDjangoOptions

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        _meta: Optional[OptimizedDjangoOptions] = None,
        model: Optional[type[Model]] = None,
        fields: Union[list[str], Literal["__all__"], None] = None,
        max_complexity: Optional[int] = None,
        **options: Any,
    ) -> None:
        if not is_valid_django_model(model):  # pragma: no cover
            msg = f"You need to pass a valid Django Model in {cls.__name__}.Meta, received {model}."
            raise TypeError(msg)

        if _meta is None:
            _meta = OptimizedDjangoOptions(cls)

        if not hasattr(cls, "pk") and (fields == ALL_FIELDS or fields is None or "pk" in fields):
            cls.pk = graphene.Int() if model._meta.pk.name == "id" else graphene.ID()
            cls.resolve_pk = cls.resolve_id

        filterset_class = options.get("filterset_class")
        filter_fields: Optional[dict[str, list[str]]] = options.pop("filter_fields", None)

        if filterset_class is None and filter_fields is not None:
            from .filter import create_filterset

            options["filterset_class"] = create_filterset(model, filter_fields)

        elif filterset_class is not None:
            from graphene_django.filter.utils import replace_csv_filters

            replace_csv_filters(filterset_class)

        _meta.max_complexity = max_complexity or optimizer_settings.MAX_COMPLEXITY
        super().__init_subclass_with_meta__(_meta=_meta, model=model, fields=fields, **options)

    @classmethod
    def pre_optimization_hook(cls, queryset: QuerySet[TModel], optimizer: QueryOptimizer) -> QuerySet[TModel]:
        """A hook for modifying the optimizer results before optimization happens."""
        return queryset

    @classmethod
    def filter_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        """Implement this method filter to the available rows from the model on this node."""
        return queryset

    @classmethod
    def get_queryset(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        return queryset

    @classmethod
    def get_node(cls, info: GQLInfo, pk: PK) -> Optional[TModel]:
        queryset = cls._meta.model._default_manager.all()
        maybe_instance = optimize_single(queryset, info, pk=pk, max_complexity=cls._meta.max_complexity)
        if maybe_instance is not None:  # pragma: no cover
            cls.run_instance_checks(maybe_instance, info)
        return maybe_instance

    @classmethod
    def run_instance_checks(cls, instance: TModel, info: GQLInfo) -> None:
        """A hook for running checks after getting a single instance."""



from __future__ import annotations

import dataclasses
from copy import copy
from typing import TYPE_CHECKING

from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Model, Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.functions import RowNumber
from graphene.utils.str_converters import to_snake_case
from graphene_django.registry import get_global_registry
from graphene_django.settings import graphene_settings

from .ast import get_model_field
from .filter_info import get_filter_info
from .prefetch_hack import _register_for_prefetch_hack
from .settings import optimizer_settings
from .typing import Generic, TModel
from .utils import (
    SubqueryCount,
    add_slice_to_queryset,
    calculate_queryset_slice,
    calculate_slice_for_queryset,
    mark_optimized,
    optimizer_logger,
    swappable_by_subclassing,
)
from .validators import validate_pagination_args

if TYPE_CHECKING:
    from .types import DjangoObjectType
    from .typing import (
        Any,
        ExpressionKind,
        GQLInfo,
        GraphQLFilterInfo,
        Literal,
        Optional,
        QuerySetResolver,
        ToManyField,
    )

__all__ = [
    "QueryOptimizer",
]


@dataclasses.dataclass
class OptimizationResults(Generic[TModel]):
    name: str | None = None
    queryset: QuerySet[TModel] | None = None
    only_fields: list[str] = dataclasses.field(default_factory=list)
    related_fields: list[str] = dataclasses.field(default_factory=list)
    select_related: list[str] = dataclasses.field(default_factory=list)
    prefetch_related: list[Prefetch | str] = dataclasses.field(default_factory=list)

    def __add__(self, other: OptimizationResults) -> OptimizationResults:
        """Adding two compilation results together means extending the lookups to the other model."""
        self.select_related.append(other.name)
        self.only_fields.extend(f"{other.name}{LOOKUP_SEP}{only}" for only in other.only_fields)
        self.related_fields.extend(f"{other.name}{LOOKUP_SEP}{only}" for only in other.related_fields)
        self.select_related.extend(f"{other.name}{LOOKUP_SEP}{select}" for select in other.select_related)

        for prefetch in other.prefetch_related:
            if isinstance(prefetch, str):
                self.prefetch_related.append(f"{other.name}{LOOKUP_SEP}{prefetch}")
            if isinstance(prefetch, Prefetch):
                prefetch.add_prefix(other.name)
                self.prefetch_related.append(prefetch)

        return self


@swappable_by_subclassing
class QueryOptimizer:
    """Creates optimized queryset based on the optimization data found by the OptimizationCompiler."""

    def __init__(
        self,
        model: type[Model] | None,
        info: GQLInfo,
        name: Optional[str] = None,
        parent: QueryOptimizer | None = None,
    ) -> None:
        self.model = model
        self.info = info
        self.only_fields: list[str] = []
        self.related_fields: list[str] = []
        self.aliases: dict[str, ExpressionKind] = {}
        self.annotations: dict[str, ExpressionKind] = {}
        self.select_related: dict[str, QueryOptimizer] = {}
        self.prefetch_related: dict[str, QueryOptimizer] = {}
        self.manual_optimizers: dict[str, QuerySetResolver] = {}
        self.total_count: bool = False
        self.name = name
        self.parent: QueryOptimizer | None = parent

    def optimize_queryset(self, queryset: QuerySet[TModel]) -> QuerySet[TModel]:
        """
        Add the optimizations in this optimizer to the given queryset.

        :param queryset: QuerySet to optimize.
        """
        filter_info = get_filter_info(self.info, queryset.model)
        results = self.process(queryset, filter_info)
        return self.optimize(results, filter_info)

    def pre_processing(self, queryset: QuerySet[TModel]) -> QuerySet[TModel]:
        """Run all pre-optimization hooks on the objct type mathcing the queryset's model."""
        object_type: Optional[DjangoObjectType] = get_global_registry().get_type_for_model(queryset.model)
        if callable(getattr(object_type, "pre_optimization_hook", None)):
            return object_type.pre_optimization_hook(queryset, self)

        return queryset  # pragma: no cover

    def process(self, queryset: QuerySet[TModel], filter_info: GraphQLFilterInfo) -> OptimizationResults[TModel]:
        """Process compiled optimizations to optimize the given queryset."""
        queryset = self.pre_processing(queryset)
        queryset = self.run_manual_optimizers(queryset, filter_info)

        results = OptimizationResults(
            name=self.name,
            queryset=queryset,
            only_fields=self.only_fields,
            related_fields=self.related_fields,
        )

        for name, optimizer in self.select_related.items():
            queryset = optimizer.model._default_manager.all()
            nested_filter_info = filter_info.get("children", {}).get(name, {})
            nested_results = optimizer.process(queryset, nested_filter_info)

            # Promote `select_related` to `prefetch_related` if any annotations are needed.
            if optimizer.annotations:
                prefetch = optimizer.process_prefetch(name, nested_results, nested_filter_info)
                results.prefetch_related.append(prefetch)
                continue

            # Otherwise extend lookups to this model.
            results += nested_results

        for name, optimizer in self.prefetch_related.items():
            # For generic foreign keys, we don't know the model, so we can't optimize the queryset.
            if optimizer.model is None:
                results.prefetch_related.append(optimizer.name)
                continue

            queryset = optimizer.model._default_manager.all()
            nested_filter_info = filter_info.get("children", {}).get(name, {})
            nested_results = optimizer.process(queryset, nested_filter_info)

            prefetch = optimizer.process_prefetch(name, nested_results, nested_filter_info)
            results.prefetch_related.append(prefetch)

        return results

    def optimize(self, results: OptimizationResults[TModel], filter_info: GraphQLFilterInfo) -> QuerySet[TModel]:
        """Optimize the given queryset based on the optimization results."""
        queryset = results.queryset

        if results.select_related:
            queryset = queryset.select_related(*results.select_related)
        if results.prefetch_related:
            queryset = queryset.prefetch_related(*results.prefetch_related)
        if not optimizer_settings.DISABLE_ONLY_FIELDS_OPTIMIZATION and (results.only_fields or results.related_fields):
            queryset = queryset.only(*results.only_fields, *results.related_fields)
        if self.aliases:
            queryset = queryset.alias(**self.aliases)
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        queryset = self.filter_queryset(queryset, filter_info)

        mark_optimized(queryset)
        return queryset

    def process_prefetch(self, to_attr: str, results: OptimizationResults, filter_info: GraphQLFilterInfo) -> Prefetch:
        """Process a prefetch, optimizing its queryset based on the given filter info."""
        queryset = self.optimize(results, filter_info)
        queryset = self.paginate_prefetch_queryset(queryset, filter_info)
        return Prefetch(self.name, queryset, to_attr=to_attr if to_attr != self.name else None)

    def paginate_prefetch_queryset(self, queryset: QuerySet, filter_info: GraphQLFilterInfo) -> QuerySet:
        """Paginate prefetch queryset based on the given filter info after it has been filtered."""
        # Only paginate nested connection fields.
        if not filter_info.get("is_connection", False):
            return queryset

        field: Optional[ToManyField] = get_model_field(self.parent.model, self.name)
        if field is None:  # pragma: no cover
            msg = (
                f"Cannot find field {self.name!r} on model {self.parent.model.__name__!r}. "
                f"Cannot optimize nested pagination."
            )
            optimizer_logger.warning(msg)
            return queryset

        remote_field = field.remote_field
        field_name = (
            remote_field.name  # .
            if isinstance(field, models.ManyToManyField | GenericRelation)
            else remote_field.attname
        )

        order_by = self.get_prefetch_ordering(filter_info, model=queryset.model)

        pagination_args = validate_pagination_args(
            after=filter_info.get("filters", {}).get("after"),
            before=filter_info.get("filters", {}).get("before"),
            offset=filter_info.get("filters", {}).get("offset"),
            first=filter_info.get("filters", {}).get("first"),
            last=filter_info.get("filters", {}).get("last"),
            max_limit=filter_info.get("max_limit", graphene_settings.RELAY_CONNECTION_MAX_LIMIT),
        )

        if (
            self.total_count
            or pagination_args.get("last") is not None
            or pagination_args.get("after") is not None
            or pagination_args.get("size") is None
        ):
            # If the query asks for total count for a nested connection field,
            # or is trying to limit the number of items from the end of the list,
            # or the user has set the `max_size` for the field to None (=no limit),
            # annotate the models in the queryset with the total count for each partition.
            # This is optional, since there is a performance impact due to needing
            # to use a subquery for each partition.
            queryset = queryset.annotate(
                **{
                    optimizer_settings.PREFETCH_COUNT_KEY: SubqueryCount(
                        queryset.filter(**{field_name: models.OuterRef(field_name)}),
                    ),
                },
            )

        # Don't limit the queryset if no pagination arguments are given (and field `max_size=None`)
        if all(value is None for value in pagination_args.values()):  # pragma: no cover
            return queryset

        if (
            pagination_args.get("last") is not None
            or pagination_args.get("after") is not None
            or pagination_args.get("size") is None
        ):
            queryset = calculate_slice_for_queryset(queryset, **pagination_args)
        else:
            cut = calculate_queryset_slice(**pagination_args)
            queryset = add_slice_to_queryset(queryset, start=models.Value(cut.start), stop=models.Value(cut.stop))

        _register_for_prefetch_hack(self.info, field)

        return (
            queryset
            # Add a row number to the queryset, and limit the rows for each
            # partition based on the given pagination arguments.
            .alias(
                **{
                    optimizer_settings.PREFETCH_PARTITION_INDEX: (
                        models.Window(
                            expression=RowNumber(),
                            partition_by=models.F(field_name),
                            order_by=order_by,
                        )
                        - models.Value(1)  # Start from zero.
                    )
                },
            ).filter(
                **{
                    f"{optimizer_settings.PREFETCH_PARTITION_INDEX}__gte": models.F(
                        optimizer_settings.PREFETCH_SLICE_START
                    ),
                    f"{optimizer_settings.PREFETCH_PARTITION_INDEX}__lt": models.F(
                        optimizer_settings.PREFETCH_SLICE_STOP
                    ),
                }
            )
        )

    def get_prefetch_ordering(self, filter_info: GraphQLFilterInfo, model: type[Model]) -> list[str]:
        """Get the ordering for prefetch querysets."""
        order_info: str | list[str] = self.process_order_by_filter(filter_info.get("filters", {}).get("order_by", ""))

        # 'Graphene-django' uses a comma-separated string for ordering.
        if isinstance(order_info, str):
            order_info = order_info.split(",")
        # 'Graphene-django-extensions' uses a list of enums.
        elif isinstance(order_info, list):  # pragma: no cover
            order_info = [str(x) for x in order_info]

        return (
            # Use the `order_by` from the filter info, if available
            [x for x in order_info if x]
            # Use the model's `Meta.ordering` if no `order_by` is given
            or copy(model._meta.ordering)
            # No ordering if neither is available
            or []
        )

    def filter_queryset(self, queryset: QuerySet, filter_info: GraphQLFilterInfo) -> QuerySet:
        """Run all filtering based on the object type matching the queryset's model."""
        # Run filtering hooks on object types if they exist.
        object_type: Optional[DjangoObjectType] = get_global_registry().get_type_for_model(queryset.model)
        if callable(getattr(object_type, "filter_queryset", None)):
            queryset = object_type.filter_queryset(queryset, self.info)

        # Lastly, run filterset filtering, if any.
        if filter_info.get("filterset_class") is None:
            return queryset

        filterset = filter_info["filterset_class"](
            data=self.process_filters(filter_info["filters"]),
            queryset=queryset,
            request=self.info.context,
        )
        if not filterset.is_valid():  # pragma: no cover
            raise ValidationError(filterset.form.errors.as_json())

        return filterset.qs

    def process_order_by_filter(self, value: list[str] | str | None) -> list[str] | str | None:
        if not value:
            return value

        if isinstance(value, list):
            return [to_snake_case(item) for item in value]

        return to_snake_case(value)

    def process_filters(self, input_data: dict[str, Any]) -> dict[str, Any]:
        from graphene_django.filter.fields import convert_enum

        kwargs = {}
        for key, value in input_data.items():
            v = convert_enum(value)

            if key == "order_by":
                v = self.process_order_by_filter(v)

            kwargs[key] = v

        return kwargs

    def run_manual_optimizers(self, queryset: QuerySet, filter_info: GraphQLFilterInfo) -> QuerySet:
        for name, func in self.manual_optimizers.items():
            filters: dict[str, Any] = filter_info.get("children", {}).get(name, {}).get("filters", {})
            queryset = func(queryset, self, **filters)
        return queryset

    def has_child_optimizer(self, name: str) -> bool:  # pragma: no cover
        return name in self.select_related or name in self.prefetch_related

    def get_child_optimizer(self, name: str) -> QueryOptimizer | None:  # pragma: no cover
        return self.select_related.get(name) or self.prefetch_related.get(name)

    def get_or_set_child_optimizer(  # pragma: no cover
        self,
        name: str,
        optimizer: QueryOptimizer,
        *,
        set_as: Literal["select_related", "prefetch_related"] = "select_related",
    ) -> QueryOptimizer:
        maybe_optimizer = self.select_related.get(name)
        if maybe_optimizer is not None:
            return maybe_optimizer
        maybe_optimizer = self.prefetch_related.get(name)
        if maybe_optimizer is not None:
            return maybe_optimizer
        getattr(self, set_as)[name] = optimizer
        return optimizer



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from django.utils.module_loading import import_string
from django_filters import FilterSet
from graphene_django.filter.utils import replace_csv_filters

from .settings import optimizer_settings

if TYPE_CHECKING:
    from django.db import models


__all__ = [
    "FilterSet",
    "create_filterset",
    "default_filterset_class",
]


@cache
def default_filterset_class() -> type[FilterSet]:
    if optimizer_settings.DEFAULT_FILTERSET_CLASS:  # pragma: no cover
        return import_string(optimizer_settings.DEFAULT_FILTERSET_CLASS)
    return FilterSet


def create_filterset(
    model: type[models.Model],
    fields: dict[str, list[str]],
) -> type[FilterSet]:
    name = f"{model._meta.object_name}FilterSet"
    meta = type("Meta", (), {"model": model, "fields": fields})
    base = default_filterset_class()
    filterset_class: type[FilterSet] = type(name, (base,), {"Meta": meta})  # type: ignore[attr-defined]
    replace_csv_filters(filterset_class)
    return filterset_class



from __future__ import annotations

import contextlib
from contextlib import suppress
from typing import TYPE_CHECKING

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Field, ForeignKey, Model
from graphene import Connection, ObjectType, PageInfo
from graphene.relay.node import AbstractNode
from graphene.types.definitions import GrapheneInterfaceType, GrapheneObjectType, GrapheneUnionType
from graphene.utils.str_converters import to_snake_case
from graphene_django import DjangoObjectType
from graphene_django.registry import get_global_registry
from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLField,
    GraphQLOutputType,
    GraphQLSchema,
    InlineFragmentNode,
    SelectionNode,
    Undefined,
)
from graphql.execution.execute import get_field_def

from .errors import OptimizerError
from .settings import optimizer_settings
from .typing import GRAPHQL_BUILTIN, GQLInfo, ModelField, Optional, ToManyField, ToOneField, TypeGuard, Union, overload

if TYPE_CHECKING:
    from graphene.types.base import BaseOptions

__all__ = [
    "GraphQLASTWalker",
]


GrapheneType = Union[GrapheneObjectType, GrapheneUnionType]
Selections = tuple[SelectionNode, ...]


class GraphQLASTWalker:
    """Class for walking the GraphQL AST and handling the different nodes."""

    def __init__(self, info: GQLInfo, model: Optional[type[Model]] = None) -> None:
        self.info = info
        self.complexity: int = 0
        self.model: type[Model] = model

    def increase_complexity(self) -> None:
        self.complexity += 1

    def run(self) -> None:
        selections: Selections = self.info.field_nodes  # type: ignore[assignment]
        field_type: GrapheneObjectType = self.info.parent_type  # type: ignore[assignment]
        return self.handle_selections(field_type, selections)

    def handle_selections(self, field_type: GrapheneType, selections: Selections) -> None:
        for selection in selections:
            if isinstance(selection, FieldNode):
                self.handle_field_node(field_type, selection)

            elif isinstance(selection, FragmentSpreadNode):
                self.handle_fragment_spread(field_type, selection)

            elif isinstance(selection, InlineFragmentNode):
                self.handle_inline_fragment(field_type, selection)

            else:  # pragma: no cover
                msg = f"Unhandled selection node: '{selection}'"
                raise OptimizerError(msg)

    def handle_field_node(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type: type[ObjectType] = field_type.graphene_type

        if self.info.parent_type == field_type:
            return self.handle_query_class(field_type, field_node)

        if issubclass(graphene_type, Connection):
            return self.handle_connection(field_type, field_node)

        if is_edge(field_type):
            return self.handle_edge(field_type, field_node)

        if issubclass(graphene_type, PageInfo):  # pragma: no cover
            return self.handle_page_info(field_type, field_node)

        if issubclass(graphene_type, ObjectType):
            return self.handle_object_type(field_type, field_node)

        if issubclass(graphene_type, AbstractNode):
            return self.handle_abstract_node(field_type, field_node)

        msg = f"Unhandled graphene type: '{graphene_type}'"  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover

    def handle_query_class(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_object_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        field_name = to_snake_case(field_node.name.value)
        if is_graphql_builtin(field_name):
            return self.handle_graphql_builtin(field_type, field_node)

        graphene_type: type[ObjectType] = field_type.graphene_type

        if issubclass(graphene_type, DjangoObjectType):
            return self.handle_model_field(field_type, field_node, field_name)
        return self.handle_plain_object_type(field_type, field_node)

    def handle_abstract_node(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type = get_global_registry().get_type_for_model(self.model._meta.concrete_model)
        field_type = self.info.schema.get_type(graphene_type._meta.name)
        return self.handle_object_type(field_type, field_node)

    def handle_graphql_builtin(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_plain_object_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_model_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field_name: str) -> None:
        model: type[Model] = field_type.graphene_type._meta.model
        field = get_model_field(model, field_name)

        if field is None:
            with self.use_model(model):
                return self.handle_custom_field(field_type, field_node)

        if not field.is_relation or is_foreign_key_id(field, field_node):
            with self.use_model(model):
                return self.handle_normal_field(field_type, field_node, field)

        if is_to_one(field):
            related_model = get_related_model(field, model)
            with self.use_model(field.model):
                return self.handle_to_one_field(field_type, field_node, field, related_model)

        if is_to_many(field):
            related_model = get_related_model(field, model)
            with self.use_model(model):
                return self.handle_to_many_field(field_type, field_node, field, related_model)

        msg = f"Unhandled field: '{field.name}'"  # pragma: no cover
        raise OptimizerError(msg)  # pragma: no cover

    def handle_custom_field(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_normal_field(self, field_type: GrapheneObjectType, field_node: FieldNode, field: Field) -> None: ...

    def handle_to_one_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
        related_model: type[Model] | None,
    ) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        self.increase_complexity()
        return self.handle_selections(graphene_type, selections)

    def handle_to_many_field(
        self,
        field_type: GrapheneObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
        related_model: type[Model] | None,
    ) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        self.increase_complexity()
        return self.handle_selections(graphene_type, selections)

    def handle_connection(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        if field_node.name.value == optimizer_settings.TOTAL_COUNT_FIELD:
            return self.handle_total_count(field_type, field_node)

        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_edge(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None:
        graphene_type = self.get_graphene_type(field_type, field_node)
        selections = get_selections(field_node)
        return self.handle_selections(graphene_type, selections)

    def handle_total_count(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_page_info(self, field_type: GrapheneObjectType, field_node: FieldNode) -> None: ...

    def handle_fragment_spread(self, field_type: GrapheneObjectType, fragment_spread: FragmentSpreadNode) -> None:
        name = fragment_spread.name.value
        fragment_definition = self.info.fragments[name]
        selections = get_selections(fragment_definition)
        return self.handle_selections(field_type, selections)

    def handle_inline_fragment(
        self,
        field_type: GrapheneUnionType | GrapheneInterfaceType,
        inline_fragment: InlineFragmentNode,
    ) -> None:
        fragment_type = get_fragment_type(field_type, inline_fragment, self.info.schema)
        graphene_options: BaseOptions = fragment_type.graphene_type._meta
        fragment_model: type[Model] | Undefined = getattr(graphene_options, "model", Undefined)

        if fragment_model != self.model:
            return None

        selections = get_selections(inline_fragment)
        return self.handle_selections(fragment_type, selections)

    def get_graphene_type(self, field_type: GrapheneObjectType, field_node: FieldNode) -> GrapheneType:
        graphql_field = get_field_def(self.info.schema, field_type, field_node)
        return get_underlying_type(graphql_field.type)

    def get_field_name(self, field_node: FieldNode) -> str:
        alias = getattr(field_node.alias, "value", None)
        return alias or to_snake_case(field_node.name.value)

    def get_related_field_name(self, related_field: ToOneField | ToManyField) -> str:  # pragma: no cover
        if hasattr(related_field, "cache_name"):  # New in Django 5.1
            return related_field.cache_name or related_field.name
        return related_field.get_cache_name() or related_field.name

    @contextlib.contextmanager
    def use_model(self, model: type[Model]) -> GraphQLASTWalker:
        orig_model = self.model
        try:
            self.model = model
            yield
        finally:
            self.model = orig_model


@overload
def get_underlying_type(
    field_type: type[GraphQLOutputType],
) -> type[Union[DjangoObjectType, GrapheneObjectType]]: ...  # pragma: no cover


@overload
def get_underlying_type(
    field_type: GraphQLOutputType,
) -> Union[DjangoObjectType, GrapheneObjectType]: ...  # pragma: no cover


def get_underlying_type(field_type):
    while hasattr(field_type, "of_type"):
        field_type = field_type.of_type
    return field_type


def get_selections(field_node: Union[FieldNode, FragmentDefinitionNode, InlineFragmentNode]) -> Selections:
    return getattr(field_node.selection_set, "selections", ())


def is_edge(field_type: GrapheneObjectType) -> bool:
    # Edge-classes are created by `graphene.relay.connection.get_edge_class`,
    # which means that we cannot check against the EdgeBase class directly.
    return all(field in field_type.fields for field in ["node", "cursor"]) and field_type.name.endswith("Edge")


def is_connection(graphql_field: Union[GraphQLField, GrapheneObjectType]) -> bool:
    return issubclass(getattr(graphql_field, "graphene_type", type(None)), Connection)


def is_node(graphql_field: GraphQLField) -> bool:
    return issubclass(getattr(getattr(graphql_field.resolve, "func", None), "__self__", type(None)), AbstractNode)


def is_graphql_builtin(field_name: str) -> bool:
    return field_name.lower() in GRAPHQL_BUILTIN


def is_foreign_key_id(field: Field, field_node: FieldNode) -> bool:
    return isinstance(field, ForeignKey) and field.get_attname() == to_snake_case(field_node.name.value)


def is_to_many(field: Field) -> TypeGuard[ToManyField]:
    return bool(field.one_to_many or field.many_to_many)


def is_to_one(field: Field) -> TypeGuard[ToOneField]:
    return bool(field.many_to_one or field.one_to_one)


def get_fragment_type(
    field_type: GrapheneUnionType | GrapheneInterfaceType,
    inline_fragment: InlineFragmentNode,
    schema: GraphQLSchema,
) -> GrapheneObjectType:
    fragment_type_name = inline_fragment.type_condition.name.value

    # For unions, fetch the type from in the union.
    if isinstance(field_type, GrapheneUnionType):
        gen = (t for t in field_type.types if t.name == fragment_type_name)
        fragment_type: Optional[GrapheneObjectType] = next(gen, None)
        if fragment_type is None:  # pragma: no cover
            msg = f"Fragment type '{fragment_type_name}' not found in union '{field_type}'"
            raise OptimizerError(msg)

    # For interfaces, fetch the type from in the schema.
    else:
        fragment_type: Optional[GrapheneObjectType] = schema.get_type(fragment_type_name)
        if fragment_type is None:  # pragma: no cover
            msg = f"Fragment type '{fragment_type_name}' not found in schema."
            raise OptimizerError(msg)

    return fragment_type


def get_related_model(related_field: Union[ToOneField, ToManyField], model: type[Model]) -> type[Model] | None:
    """
    Get the related model for a field.
    Note: For generic foreign keys, the related model is unknown (=None).
    """
    related_model = related_field.related_model
    if related_model == "self":  # pragma: no cover
        return model
    return related_model  # type: ignore[return-value]


def get_model_field(model: type[Model], field_name: str) -> Optional[ModelField]:
    if field_name == "pk":
        model_field: ModelField = model._meta.pk
        return model_field

    with suppress(FieldDoesNotExist):
        model_field: ModelField = model._meta.get_field(field_name)
        return model_field

    # Field might be a reverse many-related field without `related_name`, in which case
    # the `model._meta.fields_map` will store the relation without the "_set" suffix.
    if field_name.endswith("_set"):
        with suppress(FieldDoesNotExist):
            model_field: ModelField = model._meta.get_field(field_name.removesuffix("_set"))
            if is_to_many(model_field):
                return model_field

    return None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django.db import models

from .settings import optimizer_settings

if TYPE_CHECKING:
    from .typing import Any, Optional, ParamSpec, TypeVar, Union

    T = TypeVar("T")
    P = ParamSpec("P")
    Ttype = TypeVar("Ttype", bound=type)


__all__ = [
    "SubqueryCount",
    "add_slice_to_queryset",
    "calculate_slice_for_queryset",
    "is_optimized",
    "mark_optimized",
    "optimizer_logger",
    "remove_optimized_mark",
    "swappable_by_subclassing",
]


optimizer_logger = logging.getLogger("query_optimizer")


def mark_optimized(queryset: models.QuerySet) -> None:
    """Mark queryset as optimized so that later optimizers know to skip optimization"""
    queryset._hints[optimizer_settings.OPTIMIZER_MARK] = True


def remove_optimized_mark(queryset: models.QuerySet) -> None:  # pragma: no cover
    """Mark queryset as unoptimized so that later optimizers will run optimization"""
    queryset._hints.pop(optimizer_settings.OPTIMIZER_MARK, None)


def is_optimized(queryset: Union[models.QuerySet, list[models.Model]]) -> bool:
    """Has the queryset been optimized?"""
    # If Prefetch(..., to_attr=...) is used, the relation is a list of models.
    if isinstance(queryset, list):
        return True
    return queryset._hints.get(optimizer_settings.OPTIMIZER_MARK, False)


def calculate_queryset_slice(
    *,
    after: Optional[int],
    before: Optional[int],
    first: Optional[int],
    last: Optional[int],
    size: int,
) -> slice:
    """
    Calculate queryset slicing based on the provided arguments.
    Before this, the arguments should be validated so that:
     - `first` and `last`, positive integers or `None`
     - `after` and `before` are non-negative integers or `None`
     - If both `after` and `before` are given, `after` is less than or equal to `before`

    This function is based on the Relay pagination algorithm.
    See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm

    :param after: The index after which to start (exclusive).
    :param before: The index before which to stop (exclusive).
    :param first: The number of items to return from the start.
    :param last: The number of items to return from the end (after evaluating first).
    :param size: The total number of items in the queryset.
    """
    #
    # Start from form fetching max number of items.
    #
    start: int = 0
    stop: int = size
    #
    # If `after` is given, change the start index to `after`.
    # If `after` is greater than the current queryset size, change it to `size`.
    #
    if after is not None:
        start = min(after, stop)
    #
    # If `before` is given, change the stop index to `before`.
    # If `before` is greater than the current queryset size, change it to `size`.
    #
    if before is not None:
        stop = min(before, stop)
    #
    # If first is given, and it's smaller than the current queryset size,
    # change the stop index to `start + first`
    # -> Length becomes that of `first`, and the items after it have been removed.
    #
    if first is not None and first < (stop - start):
        stop = start + first
    #
    # If last is given, and it's smaller than the current queryset size,
    # change the start index to `stop - last`.
    # -> Length becomes that of `last`, and the items before it have been removed.
    #
    if last is not None and last < (stop - start):
        start = stop - last

    return slice(start, stop)


def calculate_slice_for_queryset(
    queryset: models.QuerySet,
    *,
    after: Optional[int],
    before: Optional[int],
    first: Optional[int],
    last: Optional[int],
    size: int,
) -> models.QuerySet:
    """
    Annotate queryset with pagination slice start and stop indexes.
    This is the Django ORM equivalent of the `calculate_queryset_slice` function.
    """
    size_key = optimizer_settings.PREFETCH_COUNT_KEY
    # If the queryset has not been annotated with the total count, add an alias with the provided size.
    # (Since this is used in prefetch QuerySets, the provided size is likely wrong though.)
    if size_key not in queryset.query.annotations:  # pragma: no cover
        queryset = queryset.alias(**{size_key: models.Value(size)})

    start = models.Value(0)
    stop = models.F(optimizer_settings.PREFETCH_COUNT_KEY)

    if after is not None:
        start = models.Case(
            models.When(
                models.Q(**{f"{size_key}__lt": after}),
                then=stop,
            ),
            default=models.Value(after),
            output_field=models.IntegerField(),
        )

    if before is not None:
        stop = models.Case(
            models.When(
                models.Q(**{f"{size_key}__lt": before}),
                then=stop,
            ),
            default=models.Value(before),
            output_field=models.IntegerField(),
        )

    if first is not None:
        queryset = queryset.alias(**{f"{size_key}_size_1": stop - start})
        stop = models.Case(
            models.When(
                models.Q(**{f"{size_key}_size_1__lt": first}),
                then=stop,
            ),
            default=start + models.Value(first),
            output_field=models.IntegerField(),
        )

    if last is not None:
        queryset = queryset.alias(**{f"{size_key}_size_2": stop - start})
        start = models.Case(
            models.When(
                models.Q(**{f"{size_key}_size_2__lt": last}),
                then=start,
            ),
            default=stop - models.Value(last),
            output_field=models.IntegerField(),
        )

    return add_slice_to_queryset(queryset, start=start, stop=stop)


def add_slice_to_queryset(
    queryset: models.QuerySet,
    *,
    start: models.Expression,
    stop: models.Expression,
) -> models.QuerySet:
    return queryset.alias(
        **{
            optimizer_settings.PREFETCH_SLICE_START: start,
            optimizer_settings.PREFETCH_SLICE_STOP: stop,
        },
    )


class SubqueryCount(models.Subquery):
    template = "(SELECT COUNT(*) FROM (%(subquery)s) _count)"
    output_field = models.BigIntegerField()


def swappable_by_subclassing(obj: Ttype) -> Ttype:
    """Makes the decorated class return the most recently created direct subclass when it is instantiated."""
    orig_init_subclass = obj.__init_subclass__

    def init_subclass(*args: Any, **kwargs: Any) -> None:
        nonlocal obj

        new_subcls: type = obj.__subclasses__()[-1]

        def new(_: type, *_args: Any, **_kwargs: Any) -> Ttype:
            return super(type, new_subcls).__new__(new_subcls)  # type: ignore[arg-type]

        obj.__new__ = new

        return orig_init_subclass(*args, **kwargs)

    obj.__init_subclass__ = init_subclass
    return obj



from __future__ import annotations

from collections.abc import Callable, Collection, Generator, Hashable, Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NamedTuple,
    Optional,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypedDict,
    TypeGuard,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.db.models import (
    Field,
    ForeignKey,
    ForeignObject,
    ForeignObjectRel,
    Manager,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    QuerySet,
)
from graphene import Argument, Dynamic
from graphene.types.structures import Structure
from graphene.types.unmountedtype import UnmountedType
from graphene_django import DjangoObjectType
from graphene_django.types import DjangoObjectTypeOptions
from graphql import GraphQLResolveInfo
from graphql_relay import ConnectionType

if TYPE_CHECKING:
    from django.contrib.auth.models import AnonymousUser, User
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
    from django.db.models.sql import Query
    from django_filters import FilterSet

    from query_optimizer.optimizer import QueryOptimizer
    from query_optimizer.validators import PaginationArgs

__all__ = [
    "GRAPHQL_BUILTIN",
    "PK",
    "Any",
    "ArgTypeInput",
    "Callable",
    "Collection",
    "ConnectionResolver",
    "Expr",
    "ExpressionKind",
    "GQLInfo",
    "Generator",
    "Generic",
    "GraphQLFilterInfo",
    "Hashable",
    "Iterable",
    "Literal",
    "ManualOptimizerMethod",
    "ModelField",
    "ModelResolver",
    "NamedTuple",
    "ObjectTypeInput",
    "OptimizedDjangoOptions",
    "Optional",
    "ParamSpec",
    "QuerySetResolver",
    "ToManyField",
    "ToOneField",
    "TypeGuard",
    "TypeVar",
    "TypedDict",
    "Union",
    "UnmountedTypeInput",
    "cast",
    "overload",
]


TModel = TypeVar("TModel", bound=Model)
PK: TypeAlias = Any
ModelField: TypeAlias = Union[Field, ForeignObjectRel, "GenericForeignKey"]
ToManyField: TypeAlias = Union["GenericRelation", ManyToManyField, ManyToOneRel, ManyToManyRel]
ToOneField: TypeAlias = Union["GenericRelation", ForeignObject, ForeignKey, OneToOneField]
AnyUser: TypeAlias = Union["User", "AnonymousUser"]
QuerySetResolver: TypeAlias = Callable[..., Union[QuerySet, Manager, None]]
ModelResolver: TypeAlias = Callable[..., Union[Model, None]]
ConnectionResolver: TypeAlias = Callable[..., ConnectionType]
ObjectTypeInput: TypeAlias = Union[
    str,
    type[DjangoObjectType],
    Callable[[], type[DjangoObjectType]],
]
UnmountedTypeInput: TypeAlias = Union[
    str,
    type[UnmountedType],
    type[DjangoObjectType],
    Structure,
    Callable[[], type[UnmountedType]],
    Callable[[], type[DjangoObjectType]],
]
ArgTypeInput: TypeAlias = Union[Argument, UnmountedType, Dynamic]
Expr: TypeAlias = Union[models.Expression, models.F, models.Q]


GRAPHQL_BUILTIN = (
    "__typename",
    "__schema",
    "__type",
    "__typekind",
    "__field",
    "__inputvalue",
    "__enumvalue",
    "__directive",
)


class UserHintedWSGIRequest(WSGIRequest):
    user: AnyUser

    optimizer_pagination: dict[str, PaginationArgs]
    """
    This attribute is only present if it was set in 'DjangoConnectionField'.
    The key is the field's name in 'snake_case'.

    The information is also not final, since the size of the slice depends on
    the number of items in the queryset after filtering.
    It is only meant to be used for optimization, if the queryset is large,
    and needs to be evaluated before the optimizer does so.
    """


class GQLInfo(GraphQLResolveInfo):
    context: UserHintedWSGIRequest


class OptimizedDjangoOptions(DjangoObjectTypeOptions):
    max_complexity: int


class GraphQLFilterInfo(TypedDict, total=False):
    name: str
    filters: dict[str, Any]
    children: dict[str, GraphQLFilterInfo]
    filterset_class: Optional[type[FilterSet]]
    is_connection: bool
    is_node: bool
    max_limit: Optional[int]


class ExpressionKind(Protocol):
    def resolve_expression(
        self,
        query: Query,
        allow_joins: bool,  # noqa: FBT001
        reuse: set[str] | None,
        summarize: bool,  # noqa: FBT001
        for_save: bool,  # noqa: FBT001
    ) -> ExpressionKind: ...


class ManualOptimizerMethod(Protocol):
    def __call__(self, queryset: QuerySet, optimizer: QueryOptimizer, **kwargs: Any) -> QuerySet: ...



from __future__ import annotations

import contextlib
from collections import defaultdict
from contextlib import nullcontext
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING
from unittest.mock import patch
from weakref import WeakKeyDictionary

from django.db import models
from django.db.models.fields.related_descriptors import _filter_prefetch_queryset

if TYPE_CHECKING:
    from graphql import OperationDefinitionNode

    from .typing import GQLInfo, TModel, ToManyField, TypeAlias

__all__ = [
    "_register_for_prefetch_hack",
    "fetch_context",
]


_PrefetchCacheType: TypeAlias = defaultdict[str, defaultdict[str, set[str]]]
_PREFETCH_HACK_CACHE: WeakKeyDictionary[OperationDefinitionNode, _PrefetchCacheType] = WeakKeyDictionary()


def _register_for_prefetch_hack(info: GQLInfo, field: ToManyField) -> None:
    # Registers the through table of a many-to-many field for the prefetch hack.
    # See `_prefetch_hack` for more information.
    if not isinstance(field, models.ManyToManyField | models.ManyToManyRel):
        return

    forward_field: models.ManyToManyField = field.remote_field if isinstance(field, models.ManyToManyRel) else field
    db_table = field.related_model._meta.db_table
    field_name = field.remote_field.name
    through = forward_field.m2m_db_table()

    # Use the `info.operation` as the key to make sure the saved values are cleared after the request ends.
    cache = _PREFETCH_HACK_CACHE.setdefault(info.operation, defaultdict(lambda: defaultdict(set)))
    cache[db_table][field_name].add(through)


def _prefetch_hack(
    queryset: models.QuerySet,
    field_name: str,
    instances: list[models.Model],
    *,
    cache: _PrefetchCacheType,
) -> models.QuerySet:
    """
    Patches the prefetch mechanism to not create duplicate joins in the SQL query.
    This is needed due to how filtering with many-to-many relations is implemented in Django,
    which creates new joins con consecutive filters for the same relation.
    See: https://docs.djangoproject.com/en/dev/topics/db/queries/#spanning-multi-valued-relationships

    For nested connection fields, a window function with a partition on the many-to-many field
    is used to limit the queryset. This adds an OUTER join for the through table of the many-to-many field.
    Then, when the prefetch queryset is filtered by `_filter_prefetch_queryset` to just the instances
    from the parent model, an INNER join is added to the through table. This creates unnecessary duplicates
    in the SQL query, which messes up the window function's partitioning. Therefore, this hack is needed
    to prevent the INNER join from being added.
    """
    #
    # `filter_is_sticky` is set here just to prevent the `used_aliases` from being cleared
    # when the queryset is cloned for filtering in `_filter_prefetch_queryset`.
    # See: `django.db.models.sql.query.Query.chain`.
    queryset.query.filter_is_sticky = True
    #
    # Add the registered through tables for a given model and field to the Query's `used_aliases`.
    # This is passed along during the filtering that happens as a part of `_filter_prefetch_queryset`,
    # until `django.db.models.sql.query.Query.join`, which has access to it with its `reuse` argument.
    # There, this should prevent the method from adding a duplicate join.
    queryset.query.used_aliases = cache[queryset.model._meta.db_table][field_name]

    return _filter_prefetch_queryset(queryset, field_name, instances)


def _hack_context(cache: _PrefetchCacheType) -> patch:
    return patch(
        f"{_filter_prefetch_queryset.__module__}.{_filter_prefetch_queryset.__name__}",
        side_effect=partial(_prefetch_hack, cache=cache),
    )


@contextlib.contextmanager
def fetch_context(info: GQLInfo) -> contextlib.AbstractContextManager:
    """Patches the prefetch mechanism if required."""
    context = nullcontext()
    if info.operation in _PREFETCH_HACK_CACHE:
        context = _hack_context(cache=deepcopy(_PREFETCH_HACK_CACHE[info.operation]))

    try:
        with context:
            yield
    finally:
        _PREFETCH_HACK_CACHE.clear()


def fetch_in_context(queryset: models.QuerySet[TModel], info: GQLInfo) -> list[TModel]:
    """Evaluates the queryset with the prefetch hack applied."""
    with fetch_context(info):
        return list(queryset)  # the database query is executed here
